// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package goissue52287 is an empty internal package.
// It exists only to work around go.dev/issue/52287 and
// can be removed after Go 1.19 stops being supported.
package goissue52287
